package org.edg.wl.dgas.dgassim.analysis;

import org.edg.wl.dgas.dgassim.analysis.util.*;
import org.edg.wl.dgas.dgassim.analysis.math.*;
import java.util.ArrayList;
import java.io.*;

/**
   analyzes a set of DGASSim output files computing mean values, mean absolut
   deviations and standard deviations.
 */
public class DGASSimResultAnalyzer {

    /**
       the default tolerance for considering time points as being contemporary
       is <code>0</code>.
     */
    public static int DEF_TIME_TOLERANCE = 0;

    /**
       the default name of the output file that conatins the mean values is
       <code>"dgassim-analysis_mean.dat"</code>.
     */
    public static String DEF_MEAN_OUTFILE = "dgassim-analysis_mean.dat";

    /**
       the default name of the output file that conatins the mean deviations is
       <code>"dgassim-analysis_meandev.dat"</code>.
     */
    public static String DEF_MEANDEV_OUTFILE = "dgassim-analysis_meandev.dat";

    /**
       the default name of the output file that conatins the standard deviation
       is <code>"dgassim-analysis_stddev.dat"</code>.
     */
    public static String DEF_STDDEV_OUTFILE = "dgassim-analysis_stddev.dat";

    /**
       the default name of the output file that conatins the standard deviation
       (relative to the mean value) is
       <code>"dgassim-analysis_relstddev.dat"</code>.
     */
    public static String DEF_RELSTDDEV_OUTFILE
	= "dgassim-analysis_relstddev.dat";

    public static void main(String[] args) {

	int timeTolerance = DEF_TIME_TOLERANCE;
 
	String meanOutFName = DEF_MEAN_OUTFILE;
	String meanDevOutFName = DEF_MEANDEV_OUTFILE;
	String stdDevOutFName = DEF_STDDEV_OUTFILE;
	String relStdDevOutFName = DEF_RELSTDDEV_OUTFILE;

	ArrayList inputFNames = new ArrayList();

	int noInFiles;

	String comment = "";

	boolean verbose = false;

	// check command line parameters:
	try {
	    for (int num = 0; num < args.length; num++) {
		if (args[num].equals("--help"))
		    throw new Exception();
		else if (args[num].startsWith("--timeTolerance")) {
		    String[] tokens = args[num].split("=");
		    timeTolerance = Integer.parseInt(tokens[1]);
		}
		else if (args[num].startsWith("--meanOutFName")) {
		    String[] tokens = args[num].split("=");
		    meanOutFName = new String(tokens[1]);
		}
		else if (args[num].startsWith("--meanDevOutFName")) {
		    String[] tokens = args[num].split("=");
		    meanDevOutFName = new String(tokens[1]);
		}
		else if (args[num].startsWith("--stdDevOutFName")) {
		    String[] tokens = args[num].split("=");
		    stdDevOutFName = new String(tokens[1]);
		}
		else if (args[num].startsWith("--relStdDevOutFName")) {
		    String[] tokens = args[num].split("=");
		    relStdDevOutFName = new String(tokens[1]);
		}
		else if (args[num].startsWith("--comment")) {
		    String[] tokens = args[num].split("=");
		    comment = new String(tokens[1]);
		}
		else if (args[num].equals("--verbose")) {
		    verbose = true;
		}
		else { // input file name
		    inputFNames.add(new String (args[num]));
		}
	    }

	    if (inputFNames.size() == 0) // no input files specified
		throw new Exception();
	}
	catch (Exception ex) {
	    usageMessage();
	    System.exit(0);
	}


	noInFiles = inputFNames.size();


	// open input files:
	BufferedReader[] inputFiles = new BufferedReader[noInFiles];
	int fNum = 0;
	try {
	    for (; fNum < noInFiles; fNum++) {
		inputFiles[fNum] =
		    new BufferedReader(new FileReader(
					      (String)inputFNames.get(fNum))
				       );
	    }
	}
	catch (Exception ex) {
	    System.err.println("Error while reading '" + inputFNames.get(fNum)
			       + "': " + ex.getMessage());
	    System.exit(1);
	}


	// read input data:
	ArrayList[] inputResultSets = new ArrayList[noInFiles];
	try {
	    ResultTimePoint result;
	    String inLine;

	    String[] lineTokens;

	    for (fNum = 0; fNum < noInFiles; fNum++) {

		inputResultSets[fNum] = new ArrayList();

		while ( (inLine = inputFiles[fNum].readLine()) != null ) {
		    if ( !inLine.startsWith("#") && !inLine.equals("\n") ) {
			lineTokens = inLine.split(" ");
			result = new ResultTimePoint();
			result.time = Integer.parseInt(lineTokens[0]);
			result.value = (double)Integer.parseInt(lineTokens[1]);
			inputResultSets[fNum].add(result);
		    }
		}

		inputFiles[fNum].close();
	    }
	}
	catch (Exception ex) {
	    System.err.println("Error while reading '" + inputFNames.get(fNum)
			       + "': " + ex.getMessage());
	    System.exit(1);
	}


	// generate and write statistical output
	try {
	    // initialize output files
	    BufferedWriter meanOutFile =
		new BufferedWriter(new FileWriter(meanOutFName));
	    BufferedWriter meanDevOutFile =
		new BufferedWriter(new FileWriter(meanDevOutFName));
	    BufferedWriter stdDevOutFile =
		new BufferedWriter(new FileWriter(stdDevOutFName));
	    BufferedWriter relStdDevOutFile =
		new BufferedWriter(new FileWriter(relStdDevOutFName));

	    // write headers
	    String outLine = "# mean (average) values\n"
		+ "# " + comment + "\n";
	    meanOutFile.write(outLine, 0, outLine.length());
	    outLine = "# mean (average) absolute deviation\n"
		+ "# " + comment + "\n";
	    meanDevOutFile.write(outLine, 0, outLine.length());
	    outLine = "# standard deviation\n"
		+ "# " + comment + "\n";
	    stdDevOutFile.write(outLine, 0, outLine.length());
	    outLine = "# relative standard deviation\n"
		+ "# " + comment + "\n";
	    relStdDevOutFile.write(outLine, 0, outLine.length());

	    // determine maximum length (due to the price adjustment threshold
	    // not all files have the same length!
	    int maxLen = 0;
	    for (fNum = 0; fNum < noInFiles; fNum++) {
		if (inputResultSets[fNum].size() > maxLen)
		    maxLen = inputResultSets[fNum].size();
	    }

	    ResultTimePoint[] resultSet = new ResultTimePoint[noInFiles];

	    int currTime = Integer.MIN_VALUE;

	    int[] currRes = new int[noInFiles];
	    for (fNum = 0; fNum < noInFiles; fNum++) 
		currRes[fNum] = 0;

	    // construct all sets:
	    while ( true ) {
		// determine next time point (roughly,
		// mean value determined later)
		int lastTime = currTime;
		currTime = Integer.MAX_VALUE;
		for (fNum = 0; fNum < noInFiles; fNum++) {
		    if (((ResultTimePoint)inputResultSets[fNum]
			 .get(currRes[fNum])).time <
			                             currTime
			&& ((ResultTimePoint)inputResultSets[fNum]
			    .get(currRes[fNum])).time >
			                            (lastTime + timeTolerance)
			)
			// smallest value that is larger than lastTime 
			// is the current one
			currTime = ((ResultTimePoint)inputResultSets[fNum]
				    .get(currRes[fNum])).time;
		}

		// check which result respects the time tolerance, if not set
		// the currRes cursor back
		int noTimePoints = 0, totalTime = 0;
		for (fNum = 0; fNum < noInFiles; fNum++) {
		    int thisTime = ((ResultTimePoint)inputResultSets[fNum]
				    .get(currRes[fNum])).time;
		    if (thisTime > (currTime + timeTolerance)) //if later point
			currRes[fNum]--; //set it back, and read the last value
		                         //again, valid for this time point
		    else if (thisTime >= (currTime - timeTolerance)) {
			// within tolerance limit
			totalTime += thisTime;
			noTimePoints++;
		    }
		}
		// determine mean value of time points within tolerance limit:
		currTime = (int)((double)totalTime/(double)noTimePoints + 0.5);

		if (verbose)
		    System.out.println("current time point: " + currTime
				       + "   (" + comment + ")");

		if (verbose) {
		    System.out.print("  result cursors: ");
		    for (fNum = 0; fNum < noInFiles; fNum++) {
			System.out.print("<" + fNum + ">=" + currRes[fNum]
					 + "  ");
		    }
		    System.out.println("");
		}


		// now we got the values valid for the current time point
		// compute statistical data:
		for (fNum = 0; fNum < noInFiles; fNum++) {
		    resultSet[fNum] = new ResultTimePoint();
		    resultSet[fNum].time = currTime;
		    resultSet[fNum].value = 
			((ResultTimePoint)inputResultSets[fNum]
			 .get(currRes[fNum])).value;
		}
		double meanValue = StatisticalAnalyzer.meanValue(resultSet);
		double meanDev = StatisticalAnalyzer.meanDeviation(resultSet);
		double stdDev =
		    StatisticalAnalyzer.standardDeviation(resultSet);
		double relStdDev = 0.0;
		if (meanValue != 0)
		    relStdDev = stdDev/meanValue;
		else if (stdDev != 0)
		    relStdDev = Double.MAX_VALUE;


		if (verbose)
		    System.out.println("  mean = " + meanValue
				       + "; mean dev. = " + meanDev
				       + "; std. dev. = " + stdDev
				       + "; rel. std. dev. = " + relStdDev);


		// write output file:
		outLine = "" + currTime + " " + meanValue + "\n";
		meanOutFile.write(outLine, 0, outLine.length());
		outLine = "" + currTime + " " + meanDev + "\n";
		meanDevOutFile.write(outLine, 0, outLine.length());
		outLine = "" + currTime + " " + stdDev + "\n";
		stdDevOutFile.write(outLine, 0, outLine.length());
		outLine = "" + currTime + " " + relStdDev + "\n";
		relStdDevOutFile.write(outLine, 0, outLine.length());

		// check whether this was the last row and
		// increment cursors, if not the last result in the set:
		boolean wasLastRow = true;
		for (fNum = 0; fNum < noInFiles; fNum++) {
		    if ( currRes[fNum] != (inputResultSets[fNum].size() - 1) ){
			currRes[fNum]++;
			wasLastRow = false;
		    }
		}
		if (wasLastRow)
		    break;

	    }

	    // Close output files
	    meanOutFile.close();
	    meanDevOutFile.close();
	    stdDevOutFile.close();
	    relStdDevOutFile.close();
	}
	catch (Exception ex) {
	    System.err.println("Error while writing output files"
			       + ": " + ex.getMessage());
	    System.exit(1);
	}
    }


    private static void usageMessage() {
	System.out.println("Usage: java DGASSimResultAnalyzer "
			   + "[--timeTolerance=<int>] [--comment=<String>]\n"
			   + "            [--meanOutFName=<String>] "
			   + "[--meanDevOutFName=<String>]\n"
			   + "            [--stdDevOutFName=<String>] "
			   + "[--relStdDevOutFName=<String>]\n"
			   + "            <inputfile> <inputfile> ... \n\n");
    }


}
