package org.edg.wl.dgas.dgassim.util;

import java.util.*;
import java.io.*;

/**
   contains the configuration information taken from a configuration file;
   maps parameter strings to their respective values (parameter-value pairs).
 */
public class Configuration {

    private String confFileName;
    private BufferedReader inFile;
    private TreeMap confMap;

    private boolean debug;

    // Constructors
    /**
       creates an instance and immediately reads the configuration information
       from the configuration file. Empty lines will be ignored. Lines
       beginning with <code>'#'</code> will be treated as comments. Each line
       that is not empty or a comment, must have the format <code>"parameter =
       value"</code>, spaces before and after <code>'='</code> are important
       and should not be omitted.
       @param fName the name of the configuration file.
    */
    public Configuration(String fName) throws IOException {

	confFileName = fName;

	inFile = new BufferedReader(new FileReader(fName));
	read();

	inFile.close();
    }

    private void read() throws IOException {

	confMap = new TreeMap();

	String currLine;
	String parameter = new String();
	String delimiter = new String();
	String token;
	StringBuffer value = new StringBuffer();

	int lineCount = 0;

	while ( (currLine=inFile.readLine()) != null ) {

	    lineCount++;

	    if ( (currLine.length() != 0) &&
		 (currLine.toCharArray()[0] != '#') &&   // not a comment
		 (currLine.toCharArray()[0] != '\n') ) { // or newline

		// the following is for JDK 1.3 ... with JDK 1.4 this is
		// MUCH easier, just use String.split("=");
		StringTokenizer stoke = new StringTokenizer(currLine);

		int count = 0;
		while (stoke.hasMoreElements()) {

		    token = stoke.nextToken();

		    switch (count) {

		    case 0:
			parameter = new String(token);
			break;
		    case 1:
			delimiter = new String(token);
			break;
		    case 2:
			value = new StringBuffer();
			value.append(token);
			break;
		    default: // count is equal or greater than 3
			value.append(" "+token);
		    }

		    count++;
		} // while ...

		if ( (count < 3) || !delimiter.equals("=") ) {
		    System.err.println("Bad configuration file, line "
				       + lineCount + ", required format: "
				       + "'parameter = value'\n>> " +currLine);
		}
		else { // right format, add to TreeMap
		    confMap.put(parameter, value.toString());
		}
	    }

	} // while ...

	Boolean db = getBooleanValue("DEBUG");
	if (db == null)
	    debug = false;
	else
	    debug = db.booleanValue();

    } // read() ...


    // add or replace a parameter manually:
    /**
       adds or replaces a parameter manually. Converts the value into a string.
       @param parameter the string that identifies the parameter.
       @param value the value to be associated with the parameter.
     */
    synchronized public void addParameter(String parameter, Object value) {
	confMap.put(parameter, value.toString());
    }

    // convert the configuration to a string
    /**
       converts the configuration information to a string.
       @return a string containing all parameter-value pairs.
     */
    synchronized public String toString() {
	return confMap.toString();
    }

    // get whole configuration:
    /**
       gets the configuration map.
       @return a reference to the <code><b>TreeMap</b></code> containing all
       parameter-value pairs.
     */
    synchronized public TreeMap getConfigurationMap() {
	return confMap;
    }

    // get single parameters:
    /**
       gets the string value associated with a specific parameter.
       @return a string containing the value.
       @param parameter the string that specifies the parameter.
     */
    synchronized public String getStringValue(String parameter) {
	String value = (String)confMap.get(parameter);

	if (value == null && debug)
	    System.err.println("Warning: no parameter " + parameter
			       + " in cofiguration file '" + confFileName
			       + "'.");
	return value;
    }

    /**
       gets the <code><b>Integer</b></code> value associated with a specific
       parameter.
       @return an <code><b>Integer</b></code> containing the value or returns
       <code>null</code> if the value cannot be converted into an
       <code><b>Integer</b></code>.
       @param parameter the string that specifies the parameter.
    */
   synchronized public Integer getIntegerValue(String parameter) {
	String value = getStringValue(parameter);
	if (value == null) return null;

	try {
	    return new Integer(Integer.parseInt(value));
	} 
	catch (NumberFormatException e) {
	    if (debug)
		System.err.println("Warning: parameter " + parameter
				   + " in configuration file '" + confFileName
				   + "' is not an integer.");
	}

	return null;
    }

    /**
       gets the <code><b>Double</b></code> value associated with a specific 
       parameter.
       @return a <code><b>Double</b></code> containing the value or returns
       <code>null</code> if the value cannot be converted into a
       <code><b>Double</b></code>.
       @param parameter the string that specifies the parameter.
    */
    synchronized public Double getDoubleValue(String parameter) {
	String value = getStringValue(parameter);
	if (value == null) return null;

	try {
	    return new Double(Double.parseDouble(value));
	} 
	catch (NumberFormatException e) {
	    if (debug)
		System.err.println("Warning: parameter " + parameter
				   + " in configuration file '" + confFileName
				   + "' is not a double.");
	}

	return null;
    }

    /**
       gets the <code><b>Boolean</b></code> value associated with a specific
       parameter.
       @return a <code><b>Boolean</b></code> containing the value or returns
       <code>null</code> if the value cannot be converted into a
       <code><b>Boolean</b></code>.
       @param parameter the string that specifies the parameter.
    */
    synchronized public Boolean getBooleanValue(String parameter) {
	String value = getStringValue(parameter);
	if (value == null) return null;

	if (value.equals("true") || value.equals("TRUE") ||
	    value.equals("True") || value.equals("1")) {

	    return new Boolean(true);
	}
	else if (value.equals("false") || value.equals("FALSE") ||
		 value.equals("False") || value.equals("0")) {

	    return new Boolean(false);
	}
	else {
	    if (debug)
		System.err.println("Warning: parameter " + parameter
				   + " in configuration file '" + confFileName
				   + "' is not boolean.");
	}

	return null;
    }

}




