package org.edg.wl.dgas.dgassim.workload.rb;

import java.util.*;
import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.fabric.*;

/**
   implements a utility module used by the
   {@link org.edg.wl.dgas.dgassim.workload.rb.ResourceBroker ResourceBroker}
   for scheduling decisions. The
   <code><b>CompletionTimeUtilityModule</b></code> simply chooses the
   {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
   that offers the shortest completion time (= queue wait time + job run time;
   measured in time units) for the given job. If a set of resources with equal
   short completion time is available the module randomly chooses one for job
   submission.
 */
public class CompletionTimeUtilityModule implements UtilityModule {

     /**
       creates an instance.
     */
   public CompletionTimeUtilityModule() {
       // nothing to do ...
   }

    /**
       selects one of the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       for job submission (see the
       {@link CompletionTimeUtilityModule general class description}). The
       {@link org.edg.wl.dgas.dgassim.workload.rb.ResourceBroker ResourceBroker}
       will first use a
       {@link org.edg.wl.dgas.dgassim.workload.rb.Matchmaker Matchmaker} to get
       a set of matching resources for the given
       {@link org.edg.wl.dgas.dgassim.Job Job}.
       @return a reference to the selected
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
       @param ces an <code><b>ArrayList</b></code> of references to all matching
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       as returned by the
       {@link org.edg.wl.dgas.dgassim.workload.rb.Matchmaker Matchmaker}.
       @param job the {@link org.edg.wl.dgas.dgassim.Job Job} for which to
       select a suitable resource.
     */
    public ComputingElement selectCE(ArrayList ces, Job job) {

	ArrayList selectedCEs = new ArrayList();
	ComputingElement ce = null;
	int minCompletionTime  = Integer.MAX_VALUE;
	int completionTime;

	for (int i = 0; i < ces.size(); i++) {

	    ce = (ComputingElement)ces.get(i);

	    // compute completion time = QWT + job run time.
	    completionTime = ce.getCurrQWT() +
		(int) ((double)job.compEnergy() / (double)ce.getCPUPower());

	    if (minCompletionTime > completionTime) {
		// found a less loaded resource:
		selectedCEs.clear();
		minCompletionTime = completionTime;
		selectedCEs.add(ce);
	    }
	    else if (minCompletionTime == completionTime) {
		// found a CE with the same low load:
		selectedCEs.add(ce);
	    }

	} // for ...

	if (selectedCEs.size() < 1) {
	    return null; // no CE found
	}

	// if we found at least one CE
	int ranIndex = (int) (((double)Math.random())*selectedCEs.size());
	return (ComputingElement) selectedCEs.get(ranIndex);
    }

}
