package org.edg.wl.dgas.dgassim.workload.rb;

import java.util.*;
import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.fabric.*;

/**
   implements a utility module used by the
   {@link org.edg.wl.dgas.dgassim.workload.rb.ResourceBroker ResourceBroker}
   for scheduling decisions. The <code><b>LeastLoadedUtilityModule</b></code>
   simply chooses the least loaded
   {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
   It may consider <i><b>absolute</b></i> load (absoulte current queue wait
   time) or <i><b>relative</b></i> load (current queue wait time divided by
   maximum queue wait time). If a set of resources with equal low load is
   available the module randomly chooses one for job submission.
 */
public class LeastLoadedUtilityModule implements UtilityModule {

    /**
       the default accuracy if considering <i><b>relative</b></i> load is
       <code>0.01</code> (one percent). Relative loads that differ for less
       then the specified accuracy are considered equal.
     */
    public static final double DEF_ACCURACY = 0.01; // 1% of default accuracy

    /**
       defines <i><b>relative</b></i> load to be considered (see the
       {@link LeastLoadedUtilityModule general class description}). The value
       of <code>RELATIVE_LOAD</code> is <code>0</code>.
     */
    public static final int RELATIVE_LOAD = 0;
    /**
       defines <i><b>absolute</b></i> load to be considered (see the
       {@link LeastLoadedUtilityModule general class description}). The value
       of <code>ABSOLUTE_LOAD</code> is <code>1</code>.
     */
    public static final int ABSOLUTE_LOAD = 1;

    private int loadType;
    private double accuracy; // for comparison of relative load

     /**
       creates an instance. If <i><b>relative</b></i> load (see the
       {@link LeastLoadedUtilityModule general class description}) is being
       considered, the accuracy will be set as specified. Relative loads that
       differ for less then the specified accuracy are considered equal.
       @param loadType
       {@link #ABSOLUTE_LOAD LeastLoadedUtilityModule.ABSOLUTE_LOAD}
       (= <code>1</code>) or
       {@link #RELATIVE_LOAD LeastLoadedUtilityModule.RELATIVE_LOAD}
       (= <code>0</code>).
       @param acc the accuracy if considering <i><b>relative</b></i> load.
       Should be between <code>0</code> (no accuracy threshold) and
       <code>1</code> (100 percent, consider all loads to be equal).
     */
   public LeastLoadedUtilityModule(int loadType, double acc) {
	this.loadType = loadType;
	accuracy = acc;
    }
    /**
       creates an instance. If <i><b>relative</b></i> load (see the
       {@link LeastLoadedUtilityModule general class description}) is being
       considered, the accuracy will be set to
       {@link #DEF_ACCURACY DEF_ACCURACY}.
       @param loadType
       {@link #ABSOLUTE_LOAD LeastLoadedUtilityModule.ABSOLUTE_LOAD}
       (= <code>1</code>) or
       {@link #RELATIVE_LOAD LeastLoadedUtilityModule.RELATIVE_LOAD}
       (= <code>0</code>).
     */
    public LeastLoadedUtilityModule(int loadType) { // default: 1% accuracy
	this(loadType, DEF_ACCURACY);
    }
    /**
       creates an instance that per default considers <i><b>absolute</b></i>
       load (see the {@link LeastLoadedUtilityModule general class description}).
     */
    public LeastLoadedUtilityModule() { // default: abosulte load
	this(ABSOLUTE_LOAD, DEF_ACCURACY);
    }

    /**
       selects one of the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       for job submission (see the
       {@link LeastLoadedUtilityModule general class description}). The
       {@link org.edg.wl.dgas.dgassim.workload.rb.ResourceBroker ResourceBroker}
       will first use a
       {@link org.edg.wl.dgas.dgassim.workload.rb.Matchmaker Matchmaker} to get
       a set of matching resources for the given
       {@link org.edg.wl.dgas.dgassim.Job Job}.
       @return a reference to the selected
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
       @param ces an <code><b>ArrayList</b></code> of references to all matching
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       as returned by the
       {@link org.edg.wl.dgas.dgassim.workload.rb.Matchmaker Matchmaker}.
       @param job the {@link org.edg.wl.dgas.dgassim.Job Job} for which to
       select a suitable resource.
     */
    public ComputingElement selectCE(ArrayList ces, Job job) {

	ArrayList selectedCEs = new ArrayList();
	ComputingElement ce = null;
	int load = Integer.MAX_VALUE;
	double relLoad = Double.MAX_VALUE;

	for (int i = 0; i < ces.size(); i++) {

	    ce = (ComputingElement)ces.get(i);

	    if (loadType == RELATIVE_LOAD) { // use relative load:
		double relCELoad = (double)ce.getCurrQWT()
		                        / (double)ce.getMaxQWT();
		if (relLoad > (relCELoad + accuracy)) {
		    // found a less loaded resource:
		    selectedCEs.clear();
		    relLoad = relCELoad;
		    selectedCEs.add(ce);
		}
		else if (relLoad <= (relCELoad + accuracy)
			 && relLoad >= (relCELoad - accuracy)) {
		    // found a CE with the same low load:
		    selectedCEs.add(ce);
		}
	    }
	    else { // use absolute load:
		if (load > ce.getCurrQWT()) {
		    // found a less loaded resource:
		    selectedCEs.clear();
		    load = ce.getCurrQWT();
		    selectedCEs.add(ce);
		}
		else if (load == ce.getCurrQWT()) {
		    // found a CE with the same low load:
		    selectedCEs.add(ce);
		}
	    }

	} // for ...

	if (selectedCEs.size() < 1) {
	    return null; // no CE found
	}

	// if we found at least one CE
	int ranIndex = (int) (((double)Math.random())*selectedCEs.size());
	return (ComputingElement) selectedCEs.get(ranIndex);
    }

}
