package org.edg.wl.dgas.dgassim.workload.rb;

import java.util.*;
import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.monitor.*;
import org.edg.wl.dgas.dgassim.fabric.*;

/**
   simulates the Resource Broker module used by the Workload Manager of the
   <i><a href="http://www.edg.org/">European DataGrid (EDG)</a></i> project.
   It is responsible for the scheduling decisions and thus for workload
   balancing. An instance of the <code><b>ResourceBroker</b></code> class
   first uses a
   {@link org.edg.wl.dgas.dgassim.workload.rb.Matchmaker Matchmaker} to
   determine the set of Grid resources
   ({@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s)
   that match a given {@link org.edg.wl.dgas.dgassim.Job Job}'s requirements
   (in terms of processing power, queue type, ...) and then passes the set of
   matching resources to a
   {@link org.edg.wl.dgas.dgassim.workload.rb.UtilityModule UtilityModule} that
   finally selects the resource that has the highest utility for the
   {@link org.edg.wl.dgas.dgassim.Job Job}.
 */
public class ResourceBroker {

    private boolean verbose;
    private boolean debug;

    private GIS gis;

    private static Matchmaker matchmaker;

    private UtilityModule utmod;

    /**
       creates an instance.
       @param gis the {@link org.edg.wl.dgas.dgassim.monitor.GIS GIS} (Grid
       Information System) instance to which all
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       and
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}s
       are registered.
       @param mmaker the
       {@link org.edg.wl.dgas.dgassim.workload.rb.Matchmaker Matchmaker} that
       is responsible for determining a set of Grid resources
       ({@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s)
       that match a given {@link org.edg.wl.dgas.dgassim.Job Job}'s requirements.
       @param utm the
       {@link org.edg.wl.dgas.dgassim.workload.rb.UtilityModule UtilityModule}
       that finally selects the resource that has the highest utility for a
       {@link org.edg.wl.dgas.dgassim.Job Job}.
       @param verbose sets the verbose flag to print information to the
       standard output.
       @param debug sets the debug flag to print additional debugging
       information to the standard output.
     */
    public ResourceBroker(GIS gis, Matchmaker mmaker, UtilityModule utm,
			  boolean verbose, boolean debug) {

	this.gis = gis;

	matchmaker = mmaker;
	utmod = utm;

	this.verbose = verbose;
	this.debug = debug;

	if (debug)
	    System.out.println("constructor of class ResourceBroker");
    }

    /**
       submits a given  to the Grid resource ({@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}) chosen by the <code><b>ResourceBroker</b></code> (if any suitable resource exists for the given {@link org.edg.wl.dgas.dgassim.Job Job}).
       @return true if the {@link org.edg.wl.dgas.dgassim.Job Job} could be submitted to a resource, otherwise false.
       @param job the {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement} that has to be submitted.
     */
    public boolean submitJob(Job job) {
	ArrayList ces =  matchmaker.getMatchingCEList(job);

	ComputingElement selectedCE = utmod.selectCE(ces, job);
	if (selectedCE != null) { // found a suitable resource
	    // feed job with price info (instead of PA database)
	    job.setJobPrice(selectedCE.getPrice());
	    // now submit job!
	    return selectedCE.queueJob(job);
	}

	return false; // no resource found
    }

}
